#include <cstdio>
#include <sys/time.h> /* gettimeofday */
#include <unistd.h>   /* usleep */
#include <termio.h>   /* termio */

#include <algorithm>

/* Captures video from a X window.
 * Requires the SHM extension to work on tolerable speeds.
 *
 * Copyright (C) 1992,2004 Bisqwit (http://iki.fi/bisqwit/)
 *
 * TODO: Create a standalone program without all parameters being
 *       hardcoded into the program.
 *
 * To compile: gcc -o xgetimage xgetimage.c -L/usr/X11R6/lib -lX11 -lXext -lXmu
 *
 * To postprocess the results, example:
 *  riffdump -nmovi -mfps=17.94 test0.avi
 *  riffdump -nmovi -mfps=17.94 test1.avi
 *  cat test{0,1}.avi | mencoder -forceidx - -ss 1 \
 *      -oac copy -audiofile ~/wine-audiolog.mp3 \
 *      -ovc copy -o xgetimage/test-short.avi
 */

#include "settings.hh"
#include "buffer.hh"
#include "msg.hh"
#include "x.hh"

typedef unsigned long long UINT64;
static UINT64 getutime(void)
{
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return (tv.tv_sec * 1000000LL) + tv.tv_usec;
}

static unsigned frames_total    = 0;
static unsigned frames_thisfile = 0;
static UINT64 start_time;
static FILE *avifp=NULL;
static FILE *encoders[MAX_NUM_FILES];
static int current_file;
static void OpenAvi(int avinumber)
{
    char CmdLine[512];
    char avifn[64];
    sprintf(avifn, "test%d.avi", avinumber);

#if USE_MENCODER
    sprintf(CmdLine,
        "animmaker -o- -f%g -"
        "|mencoder - -o %s -ovc lavc -lavcopts"
        " vbitrate=%u:keyint=%d:preme=0:dia=-2 >& /dev/null",
            (double)FPS,
            avifn,
            BITRATE,
            (int)(FPS * 3));
#else
    sprintf(CmdLine,
        "animmaker -f%g -o%s -",
            (double)FPS,
            avifn);
#endif

    encoders[avinumber] = popen(CmdLine, "w");
}
static void InitAvi()
{
    int num;
    for(num=0; num<MAX_NUM_FILES; ++num) OpenAvi(num);
    frames_thisfile = 0;
    frames_total    = 0;
    start_time      = getutime();
    current_file    = 0;
    avifp           = encoders[current_file];
}
static void DoneAvi()
{
    int num;
    FlushBuffers();
    for(num=0; num<MAX_NUM_FILES; ++num)
    {
        fflush(encoders[num]);
        pclose(encoders[num]);
    }
    avifp = NULL;
}
static void NextAvi()
{
    fprintf(stderr, "\n");
    avifp           = encoders[++current_file];
    frames_thisfile = 0;
}

static double GetVideoExcess()
{
    UINT64 time_now = getutime();
    double videopos = frames_total / (double)FPS;
    double want_pos = (time_now - start_time) / 1e6;
    return videopos - want_pos;
}

static void SendImage(unsigned char *vidbuf)
{
    static const unsigned char tgaBuf[] = {0,0,2, 0,0, 0,0, 0,  0,0, 0,0, 0,0, 0,0, 16,32};
    unsigned framesize = sizeof(tgaBuf) + width*height*2;
    unsigned opos      = sizeof(tgaBuf);
    unsigned frames_now;
#if 0
    unsigned y;
#endif
    
    unsigned char *thisframe = new unsigned char [framesize];
    
    std::copy(tgaBuf, tgaBuf + sizeof tgaBuf,
              thisframe);
    
    thisframe[12] = width&255;
    thisframe[13] = width>>8; 
    thisframe[14] = height&255;
    thisframe[15] = height>>8; 

#if 0
    for(y=height; y-->0; )
    {
        unsigned x, ipos = y*width*2;
        for(x=0; x<width; ++x)
        {
            register unsigned short pixel = vidbuf[ipos] + (vidbuf[ipos+1]<<8);

#define   RED_LOW_BIT_MASK 0x0800
#define GREEN_LOW_BIT_MASK 0x0020
#define  BLUE_LOW_BIT_MASK 0x0001
#define MAX_RED   0x1F
#define MAX_GREEN 0x3F
#define MAX_BLUE  0x1F

register unsigned char R = (((pixel/RED_LOW_BIT_MASK  ) & MAX_RED  ) * (0x20 / (MAX_RED  +1)));
register unsigned char G = (((pixel/GREEN_LOW_BIT_MASK) & MAX_GREEN) * (0x40 / (MAX_GREEN+1)));
register unsigned char B = (((pixel/BLUE_LOW_BIT_MASK ) & MAX_BLUE ) * (0x20 / (MAX_BLUE +1)));
                                                                                              
            thisframe[opos++] = (B | (G << 5)) & 255;                                         
            thisframe[opos++] = (G >> 3) | (R << 3); 

            ipos += 2;
        }
    }
#else
    std::copy(vidbuf, vidbuf + width*height*2,
              &thisframe[opos]);
#endif
    
    frames_now = 0;
    for(;;)
    {
        double excess = GetVideoExcess();
        if(excess > 0) break; /* discard frame, we've had enough already */
        
        AddWriteTask(avifp, thisframe, framesize);
        
        ++frames_total;
        ++frames_thisfile;
        ++frames_now;

        if(frames_thisfile >= FRAMES_PER_FILE)
        {
            NextAvi();
        }
    }
    
    if(!frames_now)
        MessageDiscardFrame();
    else
        MessageCapturedFrame(frames_now);

    delete[] thisframe;
}

static int kbhit()
{               
    struct termio term, back;
    int c;
    ioctl(0, TCGETA, &term);
    ioctl(0, TCGETA, &back);
    term.c_lflag &= ~(ECHO | ICANON);
    term.c_cc[VMIN] = 0; 
    ioctl(0, TCSETA, &term);
    c = getchar();
    ioctl(0, TCSETA, &back);
    return c != EOF;
}

int main(void)
{
    InitAvi();
    
    XInit();
    
    // Get 1 image to ensure it all works
    GetImage();
    
    start_time = getutime();
    
    while(!kbhit())
    {
        MoveBuffers();
        
        double excess = GetVideoExcess();
        //fprintf(stderr, "<excess %.4f>", excess);
        if(excess > 0) usleep((unsigned) (excess * 0.98e6));
        if(GetVideoExcess() > 0.95/FPS)
        {
            usleep(1000);
            continue;
        }
        
        MoveBuffers();
        unsigned char *data = GetImage();
        if(data)
        {
            SendImage(data);
            MessageFrameCounter(frames_total);
        }
        DeleteImage();
    }
    DoneAvi();
    
    MessageEnd();
    
    XDone();
    
    return 0;
}
